/*
 * yasapi_config.c
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <uxtheme.h>
#include <resources.h>

///////////////////////////////////////////////////////////////////////////////
typedef struct _PageVMT PageVMT;
typedef struct _PageTemplate PageTemplate;
typedef struct _Page Page;
typedef struct _Config Config;

///////////////////////////////////////////////////////////////////////////////
static void ConfigSet(Config *pConfig, HWND hDlg);
static void ConfigUpdateProgress(HWND hWnd, WORD wParam);

///////////////////////////////////////////////////////////////////////////////
struct _PageVMT {
  const char *name;
  DLGPROC lpDialogFunc;
  void (*Set)(Page *pPage, Config *pConfig);
  void (*Get)(Page *pPage, Config *pConfig);
  void (*Reset)(Page *pPage, Config *pConfig);
};

struct _PageTemplate {
  char *label;
  int idd;
  PageVMT *(*GetVMT)(void);
};

struct _Page {
  PageVMT *vmt;
  HWND hDlg;
};

struct _Config {
  Player *pPlayer;
  HMODULE hModule;
  HWND hDlg;
  HWND hDlgPage;
  UINT uDevice;
  int nNumPages;
  int nPage;
  PageTemplate *pTemplates;
  Page aPages[0];
};

///////////////////////////////////////////////////////////////////////////////
enum {
  PAGE_GENERAL,
  PAGE_DEVICES,
  PAGE_OPTIONS,
  PAGE_BUFFERS,
  PAGE_SHARE_MODE=PAGE_OPTIONS,
  PAGE_PULL=PAGE_OPTIONS,
  PAGE_BALANCE=PAGE_OPTIONS,
  PAGE_SRC=PAGE_OPTIONS
};

static PageVMT *GetGeneralVMT(void);
static PageVMT *GetOptionsVMT(void);
static PageVMT *GetBuffersVMT(void);

static PageTemplate gaTemplates[]={
  { "General",IDD_PAGE_GENERAL,GetGeneralVMT },
  { "Device Options",IDD_PAGE_OPTIONS,GetOptionsVMT },
  { "Device Buffers",IDD_PAGE_BUFFERS,GetBuffersVMT }
};

///////////////////////////////////////////////////////////////////////////////
static Config *ConfigNew(Player *pPlayer, HMODULE hModule)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->options.debug;
#endif // }
  Config *pConfig;
  int nNumPages;
  size_t size;

  nNumPages=(sizeof gaTemplates)/(sizeof *gaTemplates);
  size=(sizeof *pConfig)+nNumPages*(sizeof *pConfig->aPages);

  if (NULL==(pConfig=malloc(size))) {
    DMESSAGE(debug,"allocating priv");
    goto malloc;
  }

  memset(pConfig,0,sizeof *pConfig);
  pConfig->pPlayer=pPlayer;
  pConfig->hModule=hModule;
  pConfig->hDlg=NULL;
  pConfig->hDlgPage=NULL;
  pConfig->uDevice=pPlayer->uDevice;
  pConfig->nNumPages=nNumPages;
  pConfig->nPage=pPlugin->options.page<nNumPages?pPlugin->options.page:0;
  pConfig->pTemplates=gaTemplates;
  
  return pConfig;
//cleanup:
  //free(pConfig);
malloc:
  return NULL;
}

static void ConfigDelete(Config *pConfig)
{
  free(pConfig);
}

///////////////////////////////////////////////////////////////////////////////
static void SetShareMode(HWND hDlg, int eShareMode)
{
  SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_SHAREMODE),CB_SETCURSEL,
      eShareMode,0);
}

static void SetPull(HWND hDlg, int bPull)
{
  SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_PULL),CB_SETCURSEL,bPull,0);
}

static int GetShareMode(HWND hDlg)
{
  return SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_SHAREMODE),
      CB_GETCURSEL,0,0);
}

static int GetPull(HWND hDlg)
{
  return SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_PULL),
      CB_GETCURSEL,0,0);
}

static int GetAutoConvertPCM(HWND hDlg)
{
  return IsDlgButtonChecked(hDlg,IDC_CHECKBOX_AUTOCONVERT_PCM);
}

static int GetBalance(HWND hDlg)
{
  return IsDlgButtonChecked(hDlg,IDC_CHECKBOX_BALANCE);
}

////////////////
static void SyncDevicePeriod(HWND hDlg, int eShareMode, int bPull)
{
  BOOL bShare=YASAPI_SHAREMODE_SHARE==eShareMode;
  BOOL bDevicePeriod=!(bShare&&bPull);

  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_LABEL_DEVICE_PERIOD),bDevicePeriod);
  EnableWindow(GetDlgItem(hDlg,IDC_COMBOBOX_DEVICE_PERIOD),bDevicePeriod);
  EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_SHARE_SIZE),bDevicePeriod);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_SHARE_SIZE),bDevicePeriod);
}

static void SyncSRC(HWND hDlg, int eShareMode, int bAutoConvertPCM)
{
  if (YASAPI_SHAREMODE_EXCLUSIVE==eShareMode) {
    EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_AUTOCONVERT_PCM),FALSE);
    EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_SRC_DEFAULT_QUALITY),FALSE);
  }
  else {
    EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_AUTOCONVERT_PCM),TRUE);
    EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_SRC_DEFAULT_QUALITY),
        bAutoConvertPCM);
  }
}

static void SyncBalance(HWND hDlg, int bPull, int bBalance)
{

  if (bPull) {
    EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_BALANCE),FALSE);
    EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_BALANCE_START),FALSE);
    EnableWindow(GetDlgItem(hDlg,IDC_STATIC_BALANCE_START),FALSE);
  }
  else {
    EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_BALANCE),TRUE);
    EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_BALANCE_START),bBalance);
    EnableWindow(GetDlgItem(hDlg,IDC_STATIC_BALANCE_START),bBalance);
  }
}

////////////////
static void ConfigSetGlobalShareMode(Config *pConfig, int eShareMode)
{
  SetShareMode(pConfig->hDlg,eShareMode);
}

static void ConfigSetPageShareMode(Config *pConfig, int eShareMode)
{
  SetShareMode(pConfig->aPages[PAGE_SHARE_MODE].hDlg,eShareMode);
}

static void ConfigSetGlobalPull(Config *pConfig, int bPull)
{
  SetPull(pConfig->hDlg,bPull);
}

static void ConfigSetPagePull(Config *pConfig, int bPull)
{
  SetPull(pConfig->aPages[PAGE_PULL].hDlg,bPull);
}

static int ConfigGetShareMode(Config *pConfig)
{
  return GetShareMode(pConfig->aPages[PAGE_SHARE_MODE].hDlg);
}

static int ConfigGetPull(Config *pConfig)
{
  return GetPull(pConfig->aPages[PAGE_PULL].hDlg);
}

static int ConfigGetAutoConvertPCM(Config *pConfig)
{
  return GetAutoConvertPCM(pConfig->aPages[PAGE_SRC].hDlg);
}

static int ConfigGetBalance(Config *pConfig)
{
  return GetBalance(pConfig->aPages[PAGE_BALANCE].hDlg);
}

////////////////
static void ConfigSyncDevicePeriod(Config *pConfig, int eShareMode, int bPull)
{
  SyncDevicePeriod(pConfig->aPages[PAGE_BUFFERS].hDlg,eShareMode,bPull);
}

static void ConfigSyncSRC(Config *pConfig, int eShareMode, int bAutoConvertPCM)
{
  SyncSRC(pConfig->aPages[PAGE_SRC].hDlg,eShareMode,bAutoConvertPCM);
}

static void ConfigSyncBalance(Config *pConfig, int bPull, int bBalance)
{
  SyncBalance(pConfig->aPages[PAGE_BALANCE].hDlg,bPull,bBalance);
}

////////////////
static void ConfigOnDevicePeriod(Config *pConfig)
{
  int eShareMode=ConfigGetShareMode(pConfig);
  int bPull=ConfigGetPull(pConfig);

  ConfigSyncDevicePeriod(pConfig,eShareMode,bPull);
}

static void ConfigOnSRC(Config *pConfig)
{
  int eShareMode=ConfigGetShareMode(pConfig);
  int bAutoConvertPCM=ConfigGetAutoConvertPCM(pConfig);

  ConfigSyncSRC(pConfig,eShareMode,bAutoConvertPCM);
}

static void ConfigOnBalance(Config *pConfig)
{
  int bPull=ConfigGetPull(pConfig);
  int bBalance=ConfigGetBalance(pConfig);

  ConfigSyncBalance(pConfig,bPull,bBalance);
}

static void ConfigOnShareMode(Config *pConfig)
{
  int eShareMode=ConfigGetShareMode(pConfig);
  int bPull=ConfigGetPull(pConfig);
  int bAutoConvertPCM=ConfigGetAutoConvertPCM(pConfig);

  ConfigSyncSRC(pConfig,eShareMode,bAutoConvertPCM);
  ConfigSyncDevicePeriod(pConfig,eShareMode,bPull);
}

static void ConfigOnPull(Config *pConfig)
{
  int bPull=ConfigGetPull(pConfig);
  int bBalance=ConfigGetBalance(pConfig);
  int eShareMode=ConfigGetShareMode(pConfig);

  ConfigSyncBalance(pConfig,bPull,bBalance);
  ConfigSyncDevicePeriod(pConfig,eShareMode,bPull);
}

///////////////////////////////////////////////////////////////////////////////
static void GeneralSyncVisualization(HWND hDlg, Config *pConfig)
{
  Plugin *pPlugin=pConfig->pPlayer->pPlugin;
  BOOL bVisualization=IsDlgButtonChecked(hDlg,IDC_CHECKBOX_VISUALIZATION);
  HWND hWndRing=GetDlgItem(pConfig->hDlg,IDC_PROGRESSBAR_RING);
  HWND hWndShared=GetDlgItem(pConfig->hDlg,IDC_PROGRESSBAR_SHARED);

  if (!bVisualization) {
    ConfigUpdateProgress(hWndRing,0);
    ConfigUpdateProgress(hWndShared,0);
  }

  EnableWindow(GetDlgItem(pConfig->hDlg,IDC_STATIC_RING),bVisualization);
  EnableWindow(hWndRing,bVisualization);
  EnableWindow(GetDlgItem(pConfig->hDlg,IDC_STATIC_SHARED),bVisualization);
  EnableWindow(hWndShared,bVisualization);

  pPlugin->options.bVisualization=bVisualization;
}

#if defined (YASAPI_DEBUG) // {
static void GeneralSyncDebug(Config *pConfig, HWND hDlg)
{
  Plugin *pPlugin=pConfig->pPlayer->pPlugin;
  LRESULT debug=SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_DEBUG),
      CB_GETCURSEL,0,0);

  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_LABEL_SLEEP),debug);
  EnableWindow(GetDlgItem(hDlg,IDC_COMBOBOX_SLEEP),debug);

  if (!pPlugin->options.debug&&debug)
    ConsoleOpen();
  else if (pPlugin->options.debug&&!debug)
    ConsoleClose();

  pPlugin->options.debug=debug;
}
#endif // }

static INT_PTR CALLBACK GeneralProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Config *pConfig=(Config *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    pConfig=(Config *)lParam;
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    ControlInit(gcaPluginControlList,hDlg,pConfig->hModule);
#if ! defined (YASAPI_DEBUG) // {
    EnableWindow(GetDlgItem(hDlg,IDC_STATIC_LABEL_DEBUG),FALSE);
    SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_DEBUG),
        CB_ADDSTRING,0,(LPARAM)L"Off");
    SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_DEBUG),
        CB_SETCURSEL,0,0);
    EnableWindow(GetDlgItem(hDlg,IDC_COMBOBOX_DEBUG),FALSE);

    EnableWindow(GetDlgItem(hDlg,IDC_STATIC_LABEL_SLEEP),FALSE);
    SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_SLEEP),
        CB_ADDSTRING,0,(LPARAM)L"0 sec");
    SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_SLEEP),
        CB_SETCURSEL,0,0);
    EnableWindow(GetDlgItem(hDlg,IDC_COMBOBOX_SLEEP),FALSE);
#endif // }
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDC_CHECKBOX_VISUALIZATION:
      if (BN_CLICKED==HIWORD(wParam))
        GeneralSyncVisualization(hDlg,pConfig);

      return TRUE;
#if defined (YASAPI_DEBUG) // {
    case IDC_COMBOBOX_DEBUG:
      GeneralSyncDebug(pConfig,hDlg);
#endif // }
    default:
      break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}

static void GeneralSetOptions(Page *pPage, Config *pConfig,
    const PluginOptions *pOptions)
{
  ControlSet(gcaPluginControlList,pPage->hDlg,pOptions);
#if defined (YASAPI_DEBUG) // {
  GeneralSyncDebug(pConfig,pPage->hDlg);
#endif // }
}

static void GeneralSet(Page *pPage, Config *pConfig)
{
  PluginOptions *pOptions=&pConfig->pPlayer->pPlugin->options;

  GeneralSetOptions(pPage,pConfig,pOptions);
}

static void GeneralGet(Page *pPage, Config *pConfig)
{
  PluginOptions *pOptions=&pConfig->pPlayer->pPlugin->options;

  ControlGet(gcaPluginControlList,pPage->hDlg,pOptions);
}

static void GeneralReset(Page *pPage, Config *pConfig)
{
  GeneralSetOptions(pPage,pConfig,PluginOptionsDefault());
}

static PageVMT *GetGeneralVMT(void)
{
  static PageVMT vmt;

  if (NULL==vmt.name) {
    vmt.name="General";
    vmt.lpDialogFunc=GeneralProc;
    vmt.Set=GeneralSet;
    vmt.Get=GeneralGet;
    vmt.Reset=GeneralReset;
  }

  return &vmt;
}

///////////////////////////////////////////////////////////////////////////////
static INT_PTR CALLBACK OptionsProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Config *pConfig=(Config *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    pConfig=(Config *)lParam;
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    ControlInit(gcaDeviceControlList,hDlg,pConfig->hModule);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDC_CHECKBOX_AUTOCONVERT_PCM:
    case IDC_CHECKBOX_SRC_DEFAULT_QUALITY:
      ConfigOnSRC(pConfig);
      return TRUE;
    case IDC_CHECKBOX_BALANCE:
      ConfigOnBalance(pConfig);
      return TRUE;
    case IDC_COMBOBOX_SHAREMODE:
      ConfigSetGlobalShareMode(pConfig,GetShareMode(hDlg));
      ConfigOnShareMode(pConfig);
      return TRUE;
    case IDC_COMBOBOX_PULL:
      ConfigSetGlobalPull(pConfig,GetPull(hDlg));
      ConfigOnPull(pConfig);
      return TRUE;
    default:
      break;
    }

    break;
  case WM_HSCROLL:
    ControlSync(gcaDeviceControlList,hDlg);
    break;
  default:
    break;
  }

  return FALSE;
}

static void OptionsSetOptions(Page *pPage, Config *pConfig,
    const ChannelOptions *pOptions)
{
  ControlSet(gcaDeviceControlList,pPage->hDlg,pOptions);
  ConfigOnSRC(pConfig);
  ConfigOnBalance(pConfig);
  ConfigOnDevicePeriod(pConfig);
  ConfigSetGlobalShareMode(pConfig,GetShareMode(pPage->hDlg));
  ConfigSetGlobalPull(pConfig,GetPull(pPage->hDlg));
}

static void OptionsSet(Page *pPage, Config *pConfig)
{
  Player *pPlayer=pConfig->pPlayer;
  Channel *pChannel=pPlayer->pChannel+pConfig->uDevice;
  ChannelOptions *pOptions=&pChannel->options;

  SetWindowTextW(GetDlgItem(pPage->hDlg,IDC_STATIC_DEVICE_NAME),
      pChannel->vName.pwszVal);

  OptionsSetOptions(pPage,pConfig,pOptions);
}

static void OptionsGet(Page *pPage, Config *pConfig)
{
  Player *pPlayer=pConfig->pPlayer;
  Channel *pChannel=pPlayer->pChannel+pConfig->uDevice;
  ChannelOptions *pOptions=&pChannel->options;

  ControlGet(gcaDeviceControlList,pPage->hDlg,pOptions);
}

static void OptionsReset(Page *pPage, Config *pConfig)
{
  OptionsSetOptions(pPage,pConfig,ChannelOptionsDefault());
}

static PageVMT *GetOptionsVMT(void)
{
  static PageVMT vmt;

  if (NULL==vmt.name) {
    vmt.name="Options";
    vmt.lpDialogFunc=OptionsProc;
    vmt.Set=OptionsSet;
    vmt.Get=OptionsGet;
    vmt.Reset=OptionsReset;
  }

  return &vmt;
}

///////////////////////////////////////////////////////////////////////////////
static INT_PTR CALLBACK BuffersProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Config *pConfig=(Config *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    pConfig=(Config *)lParam;
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    ControlInit(gcaBuffersControlList,hDlg,pConfig->hModule);
    return TRUE;
  case WM_HSCROLL:
    ControlSync(gcaBuffersControlList,hDlg);
    break;
  default:
    break;
  }

  return FALSE;
}

static void BuffersSet(Page *pPage, Config *pConfig)
{
  Player *pPlayer=pConfig->pPlayer;
  Channel *pChannel=pPlayer->pChannel+pConfig->uDevice;
  ChannelOptions *pOptions=&pChannel->options;

  SetWindowTextW(GetDlgItem(pPage->hDlg,IDC_STATIC_DEVICE_NAME),
      pChannel->vName.pwszVal);
  ControlSet(gcaBuffersControlList,pPage->hDlg,pOptions);
}

static void BuffersGet(Page *pPage, Config *pConfig)
{
  Player *pPlayer=pConfig->pPlayer;
  Channel *pChannel=pPlayer->pChannel+pConfig->uDevice;
  ChannelOptions *pOptions=&pChannel->options;

  ControlGet(gcaBuffersControlList,pPage->hDlg,pOptions);
}

static void BuffersReset(Page *pPage, Config *pConfig)
{
  ControlSet(gcaBuffersControlList,pPage->hDlg,ChannelOptionsDefault());
}

static PageVMT *GetBuffersVMT(void)
{
  static PageVMT vmt;

  if (NULL==vmt.name) {
    vmt.name="Buffers";
    vmt.lpDialogFunc=BuffersProc;
    vmt.Set=BuffersSet;
    vmt.Get=BuffersGet;
    vmt.Reset=BuffersReset;
  }

  return &vmt;
}

///////////////////////////////////////////////////////////////////////////////
static void ConfigOnSelChangeTabCtrl(HWND hDlg, Config *pConfig, HWND hWndTab)
{
  int debug=pConfig->pPlayer->pPlugin->options.debug;
  int nPage=TabCtrl_GetCurSel(hWndTab);
  Page *pPage=pConfig->aPages+nPage;

  if (pConfig->hDlgPage) {
    ShowWindow(pConfig->hDlgPage,FALSE);
    pConfig->hDlgPage=NULL;
  }

  pConfig->nPage=nPage;
  pConfig->hDlgPage=pPage->hDlg;
  ShowWindow(pConfig->hDlgPage,TRUE);
}

static int ConfigInitPage(HWND hDlg, HWND hWndTab, Config *pConfig, int nPage)
{
  int code=-1;
  PageTemplate *pTemplate=pConfig->pTemplates+nPage;
  Page *pPage=pConfig->aPages+nPage;
#if defined (YASAPI_DEBUG) // {
  int debug=pConfig->pPlayer->pPlugin->options.debug;
#endif // }
  TCITEMA tie;
  LPCTSTR lpTemplateName;
  RECT rc;

  pPage->vmt=pTemplate->GetVMT();

  memset(&tie,0,sizeof tie);
  tie.mask=TCIF_TEXT|TCIF_IMAGE;
  tie.iImage=-1;
  tie.pszText=pTemplate->label;
  SendMessageA(hWndTab,TCM_INSERTITEMA,nPage,(LPARAM)&tie);

  // Finding resource.
  lpTemplateName=MAKEINTRESOURCEA(pTemplate->idd);

  pPage->hDlg=CreateDialogParamA(
    pConfig->hModule,   // _In_opt_ HINSTANCE hInstance,
    lpTemplateName,   // _In_     LPCTSTR   lpTemplateName,
    hDlg,             // _In_opt_ HWND      hWndParent,
    pPage->vmt->lpDialogFunc,
                      // _In_opt_ DLGPROC   lpDialogFunc,
    (LPARAM)pConfig     // _In_     LPARAM    dwInitParam
  );

  if (NULL==pPage->hDlg) {
    DMESSAGE(debug,"creating child dialog");
    goto child;
  }

  GetClientRect(pPage->hDlg,&rc);
  TabCtrl_AdjustRect(hWndTab,FALSE,&rc);

  MapWindowPoints(
    hWndTab,          // _In_    HWND    hWndFrom,
    hDlg,             // _In_    HWND    hWndTo,
    (LPPOINT)&rc,     // _Inout_ LPPOINT lpPoints,
    2                 // _In_    UINT    cPoints
  );

  SetWindowPos(pPage->hDlg,HWND_TOP,rc.left,rc.top,0,0,
      SWP_NOSIZE|SWP_HIDEWINDOW);
  code=0;
child:
lock:
load:
find:
  return code;
}

static int ConfigInitTab(HWND hDlg, int idc, Config *pConfig)
{
  int code=-1;
#if defined (YASAPI_DEBUG) // {
  int debug=pConfig->pPlayer->pPlugin->options.debug;
#endif // }
  HWND hWndTab;
  int nPage;

  // Get the tabcontrol handle.
  if (NULL==(hWndTab=GetDlgItem(hDlg,idc))) {
    DMESSAGE(debug,"getting tabcontrol handle");
    goto item;
  }

  // Initialize pages.
  for (nPage=0;nPage<pConfig->nNumPages;++nPage) {
    if (ConfigInitPage(hDlg,hWndTab,pConfig,nPage)<0) {
      DMESSAGE(debug,"initializing page");
      goto param;
    }

#if 1 // {
    EnableThemeDialogTexture(
      pConfig->aPages[nPage].hDlg,  // _In_ HWND  hwnd,
      ETDT_ENABLETAB                // _In_ DWORD dwFlags
    );
#endif // }
  }

  SendMessageA(hWndTab,TCM_SETCURSEL,pConfig->nPage,0);
  ConfigOnSelChangeTabCtrl(hDlg,pConfig,hWndTab);
  code=0;
param:
item:
  return code;
}

static void ConfigSet(Config *pConfig, HWND hDlg)
{
  int nPage;
  Page *pPage;

  Channel *pChannel=pConfig->pPlayer->pChannel+pConfig->uDevice;
  ChannelOptions *pOptions=&pChannel->options;

  for (nPage=0;nPage<pConfig->nNumPages;++nPage) {
    pPage=pConfig->aPages+nPage;
    pPage->vmt->Set(pPage,pConfig);
  }
}

static void ConfigGet(Config *pConfig, HWND hDlg)
{
  Player *pPlayer=pConfig->pPlayer;
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->options.debug;
#endif // }
  Channel *pChannel;
  int nPage;
  Page *pPage;
  RECT rc;

  // critical section /////////////////////////////////////////////////////////
  QueueLockMutex(&pPlugin->queue); // {

  // get the coosen device/player.
  pPlayer->uDevice=pConfig->uDevice;
  pChannel=pPlayer->pChannel+pPlayer->uDevice;
  wcsncpy(pPlugin->options.id,pChannel->pstrId,YASAPI_ID_SIZE);
  // get the current tabcontrol page.
  pPlugin->options.page=pConfig->nPage;

  // read out the device/player options.
  for (nPage=0;nPage<pConfig->nNumPages;++nPage) {
    pPage=pConfig->aPages+nPage;
    pPage->vmt->Get(pPage,pConfig);
  }

  QueueUnlockMutex(&pPlugin->queue); // }
  /////////////////////////////////////////////////////////////////////////////

  GetWindowRect(hDlg,&rc);
  pPlugin->options.posx=rc.left;
  pPlugin->options.posy=rc.top;

  // save the plugin options.
  PluginOptionsSave(&pPlugin->options);
  // save the device/player options.
  ChannelOptionsSave(&pChannel->options);

#if 0 // {
#if defined (YASAPI_DEBUG) // {
  if (debug!=pPlugin->options.debug) {
    if (pPlugin->options.debug)
      ConsoleOpen();
    else
      ConsoleClose();
  }
#endif // }
#endif // }
}

static void ConfigReset(Config *pConfig, HWND hDlg)
{
  int nPage;
  Page *pPage;

  for (nPage=0;nPage<pConfig->nNumPages;++nPage) {
    pPage=pConfig->aPages+nPage;
    pPage->vmt->Reset(pPage,pConfig);
  }
}

static void ConfigEndDialog(HWND hDlg, INT_PTR nResult, Config *pConfig)
{
  Plugin *pPlugin=pConfig->pPlayer->pPlugin;

  QueueLockMutex(&pPlugin->queue); // {
  pPlugin->hDlgConfig=NULL;
  QueueUnlockMutex(&pPlugin->queue); // }
  EndDialog(hDlg,nResult);
}

static void ConfigUpdateProgress(HWND hWnd, WORD wParam)
{
  LRESULT lMin=SendMessageW(hWnd,PBM_GETRANGE,TRUE,0);
  LRESULT lMax=SendMessageW(hWnd,PBM_GETRANGE,FALSE,0);
  WORD wPos=lMin+MulDiv(lMax-lMin,wParam,USHRT_MAX);

  SendMessageW(hWnd,PBM_SETPOS,wPos,0);
}

static void ConfigSaveGUIOptions(HWND hDlg, Config *pConfig)
{
  PluginOptions *pOptions=&pConfig->pPlayer->pPlugin->options;
  const wchar_t *group=YASAPI_GROUP_GENERAL;
  const wchar_t *path=pOptions->path;
  RECT rc;

  GetWindowRect(hDlg,&rc);
  pOptions->posx=rc.left;
  pOptions->posy=rc.top;
  pOptions->page=pConfig->nPage;

  PropertySaveInt(group,YASAPI_PROPERTY_VIS,pOptions->bVisualization,path);
  PropertySaveInt(group,YASAPI_PROPERTY_POSX,pOptions->posx,path);
  PropertySaveInt(group,YASAPI_PROPERTY_POSY,pOptions->posy,path);
  PropertySaveInt(group,YASAPI_PROPERTY_PAGE,pOptions->page,path);
#if defined (YASAPI_DEBUG) // {
  PropertySaveInt(group,YASAPI_PROPERTY_DEBUG,pOptions->debug,path);
#endif // }
}

static void ConfigSyncPull(HWND hDlg, Config *pConfig)
{
  LRESULT bPull=SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_PULL),
      CB_GETCURSEL,0,0);
  HWND hDlgOptions=pConfig->aPages[PAGE_OPTIONS].hDlg;
  HWND hWndPull=GetDlgItem(hDlgOptions,IDC_COMBOBOX_PULL);

  SendMessageW(hWndPull,CB_SETCURSEL,bPull,0);
  OptionsSyncPull(hDlgOptions,pConfig,FALSE);
}

static void ConfigFillList(HWND hDlg, Config *pConfig, int idc)
{
  Player *pPlayer=pConfig->pPlayer;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->options.debug;
#endif // }
  UINT cDevices=pPlayer->cDevices,uDevice;
  Channel *pChannel=pPlayer->pChannel;
  HWND hWndCtl;

  if (NULL==(hWndCtl=GetDlgItem(hDlg,idc))) {
    DMESSAGE(debug,"getting list");
    goto list;
  }

  for (uDevice=0;uDevice<cDevices;++uDevice) {
    SendMessageW(hWndCtl,CB_ADDSTRING,0,
        (LPARAM)pChannel[uDevice].vName.pwszVal);
  }

  SendMessageW(hWndCtl,CB_SETCURSEL,pConfig->uDevice,0);
list:
  return;
}

static INT_PTR ConfigOnInit(HWND hDlg, Config *pConfig)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pConfig->pPlayer->pPlugin->options.debug;
#endif // }
  HWND hWndTip;

  pConfig->hDlg=hDlg;
  SetWindowLongPtr(hDlg,GWLP_USERDATA,(LONG_PTR)pConfig);
  SetWindowTextA(hDlg,PI_LABEL " Settings");

  if (0<=pConfig->pPlayer->pPlugin->options.posx
      &&0<=pConfig->pPlayer->pPlugin->options.posy) {

    SetWindowPos(
      hDlg,     // _In_     HWND hWnd,
      NULL,     // _In_opt_ HWND hWndInsertAfter,
      pConfig->pPlayer->pPlugin->options.posx,
                // _In_     int  X,
      pConfig->pPlayer->pPlugin->options.posy,
                // _In_     int  Y,
      0,        // _In_     int  cx,
      0,        // _In_     int  cy,
      SWP_NOOWNERZORDER|SWP_NOSIZE
                //_In_     UINT uFlags
    );
  }

  ConfigFillList(hDlg,pConfig,IDC_COMBOBOX_DEVICE);
  hWndTip=ControlInit(gcaReducedDeviceControlList,hDlg,pConfig->hModule);
  ControlAddToolTip(hDlg,hWndTip,IDC_COMBOBOX_DEVICE,
      L"The device where the audio should be routed.");

  if (ConfigInitTab(hDlg,IDC_TABCONTROL,pConfig)<0) {
    DMESSAGE(debug,"initializing tabcontrol");
    return FALSE;
  }

  ConfigSet(pConfig,hDlg);
  pConfig->pPlayer->pPlugin->hDlgConfig=hDlg;

  return TRUE;
}

static void ConfigOnSelChangeComboBox(HWND hDlg, Config *pConfig, HWND hWndCtl)
{
  pConfig->uDevice=SendMessageW(hWndCtl,CB_GETCURSEL,0,0);
  ConfigSet(pConfig,pConfig->hDlg);
}

static INT_PTR CALLBACK ConfigProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Config *pConfig=(Config *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    return ConfigOnInit(hDlg,(Config *)lParam);
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDOK:
      ConfigGet(pConfig,hDlg);
      ConfigEndDialog(hDlg,TRUE,pConfig);
      return TRUE;
    case IDCANCEL:
      ConfigSaveGUIOptions(hDlg,pConfig);
      ConfigEndDialog(hDlg,FALSE,pConfig);
      return TRUE;
    case IDC_BUTTON_SAVE:
      ConfigGet(pConfig,hDlg);
      return TRUE;
    case IDC_BUTTON_DEFAULT:
      ConfigReset(pConfig,hDlg);
      return TRUE;
    case IDC_COMBOBOX_SHAREMODE:
      ConfigSetPageShareMode(pConfig,GetShareMode(hDlg));
      ConfigOnShareMode(pConfig);
      return TRUE;
    case IDC_COMBOBOX_PULL:
      ConfigSetPagePull(pConfig,GetPull(hDlg));
      ConfigOnPull(pConfig);
      return TRUE;
    case IDC_COMBOBOX_DEVICE:
      ConfigOnSelChangeComboBox(hDlg,pConfig,(HWND)lParam);
      return TRUE;
    default:
      break;
    }

    break;
  case WM_NOTIFY:
    switch (((NMHDR *)lParam)->idFrom) {
    case IDC_TABCONTROL:
      switch (((NMHDR *)lParam)->code) {
      case TCN_SELCHANGE:
        ConfigOnSelChangeTabCtrl(hDlg,pConfig,((NMHDR *)lParam)->hwndFrom);
        return TRUE;
      default:
        break;
      }

      break;
    default:
      break;
    }

    break;
  case WM_CONFIG_UPDATE:
    if (IsDlgButtonChecked(pConfig->aPages[PAGE_GENERAL].hDlg,
        IDC_CHECKBOX_VISUALIZATION)) {
      ConfigUpdateProgress(GetDlgItem(hDlg,IDC_PROGRESSBAR_RING),
          LOWORD(lParam));
      ConfigUpdateProgress(GetDlgItem(hDlg,IDC_PROGRESSBAR_SHARED),
          HIWORD(lParam));
    }

    return TRUE;
  default:
    break;
  }

  return FALSE;
}

void ConfigDialog(HWND hWndParent, HINSTANCE hInstance, Player *pPlayer)
{
  LPCWSTR lpTemplateName=MAKEINTRESOURCEW(IDD_CONFIG);
  Config *pConfig;

  if (NULL==(pConfig=ConfigNew(pPlayer,hInstance)))
    goto config;

  DialogBoxParamW(
    hInstance,            // _In_opt_  HINSTANCE hInstance,
    lpTemplateName,       // _In_      LPCTSTR lpTemplateName,
    hWndParent,           // _In_opt_  HWND hWndParent,
    ConfigProc,           // _In_opt_  DLGPROC lpDialogFunc,
    (LPARAM)pConfig       // _In_      LPARAM dwInitParam
  );

  ConfigDelete(pConfig);
config:
  return;
}
